#ifndef _MAILBOX_H
#define _MAILBOX_H 1

#include "wordlist.h"
#include "var.h"
#include "mailerthread.h"
#include "semaphore.h"
#include "socketclasses.h"

#define ENVELOPE_I_READ			0
#define ENVELOPE_I_PARENTID		1
#define ENVELOPE_MAILFROM		2
#define ENVELOPE_NAMEFROM		3
#define ENVELOPE_SUBJECT		4
#define ENVELOPE_DATE			5
#define ENVELOPE_HOSTS			6
#define ENVELOPE_NEWSGROUPS		7
#define ENVELOPE_GROUPS			7
#define ENVELOPE_RECEIVETIME	8

//==========================================================================================
// ITCMessage/ITCPipe
//
// Relatively simple FIFO buffers to be used for communication between the GUI thread and the
// backend threads.
//==========================================================================================

// defines for ui-to-collector communication

#define CLASS_CHECKGROUP	0x10000001 // [1] arg1 = groupname
#define CLASS_GETBODY		0x10000002 // [1] arg1 = intmsgid
#define CLASS_DELETE		0x10000003 // [1] arg1 = intmsgid
#define CLASS_MOVE			0x10000004 // [3] arg1 = intmsgid    arg2 = foldername
#define CLASS_OPENSERVER	0x10000005 // [0]
#define CLASS_CLOSESERVER	0x10000006 // [0]
#define CLASS_EXPIRE		0x10000007 // [0]		
#define CLASS_DIE			0x100000FF // [0]

// defines for collector-to-ui communication

#define CLASS_CHECKEDGROUP	0x20000001 // [1] arg1 = groupname
#define CLASS_NEWARTICLE	0x20000002 // [2] arg1 = groupname   arg2 = intmsgid
#define CLASS_REMARTICLE	0x20000003 // [2] arg1 = groupname   arg2 = intmsgid
#define CLASS_ERROR			0x20000004 // [2] arg1 = subsystem   arg2 = error msg
#define CLASS_STATUS		0x20000005 // [1] arg1 = status msg
#define CLASS_GOTBODY		0x20000006 // [1] arg1 = msgid

// defines for scheduler events

#define SCHED_SAVEBOX		0x30000001
#define SCHED_CHECKNEW		0x30000002
#define SCHED_EXPIRE		0x30000003
#define SCHED_CLEANGARBAGE	0x30000004



#define SUB_ERROR_CRITICAL  0x20000001 // critical error message
#define SUB_ERROR_LIMIT		0x20000002 // operation cannot be performed due to limitations of the underlying protocol
#define SUB_ERROR_INFORM	0x20000003 // informative message about a non-critical error event

// global defines

#define NO_SUBCLASS			0x00000000 // placeholder

class ITCMessage
{
	public:
						 ITCMessage (unsigned long, unsigned long, WordList *);
						~ITCMessage (void);
		
		ITCMessage		*Next, *Prev;
		unsigned long	 Class;
		unsigned long	 Subclass;
		time_t			 Scheduled;
		int				 Reschedule;
		WordList		*Message;
};

class ITCPipe
{
	public:
						 ITCPipe (void);
						~ITCPipe (void);
		
		ITCMessage		*GetNextMessage (void);
		ITCMessage		*GetNextMessage (unsigned long);
		void			 PostMessage (unsigned long, unsigned long, WordList *);
		void			 PostMessage (unsigned long, unsigned long, char *);
		void			 PostMessage (unsigned long, unsigned long, char *, char *);
		void			 PostMessage (unsigned long, unsigned long, char *, char *, char *);
		int				 Queue (void);
	
	private:
		ASemaphore		*Lock;
		ITCMessage		*First, *Last;
		int				 Count;
};

//==========================================================================================
// Scheduler
//
// This class will be used by the "collector" thread connected to every mailbox. Repeating events,
// like an automatic mail check, are scheduled here. Future events can be scheduled once, or as
// repeating. Events are stored as ITCMessages, allowing for the same handler to be used.
//==========================================================================================

class Scheduler
{
	public:
							 Scheduler (void);
							~Scheduler (void);
		
		void				 Schedule (int, int, unsigned long, unsigned long, WordList *);
		void				 Schedule (int, int, unsigned long, unsigned long, char *);
		void				 Schedule (int, int, unsigned long, unsigned long, char *, char *);
		void				 Schedule (int, int, unsigned long, unsigned long, char *, char *, char *);
		ITCMessage			*CheckSchedule (void);
		
		int				 	 Queue (void);
		ITCMessage			*Get (int);
		void				 Remove (int);
	
	private:
		ASemaphore			*Lock;
		ITCMessage			*First, *Last;
		int					 Count;
};

//==========================================================================================
// MailViewNode
//
// An entity that represents a message as part of a folder (as opposed to a message as part of a
// mailbox, which is where this class is aimed at).
//==========================================================================================

class MailViewNode
{
	public:
							 MailViewNode (class MailMessage *, class MailView *);
							~MailViewNode (void);
		
		MailViewNode		*Next, *Prev;
		
		class MailMessage	*Message;
		class MailView		*Parent;
};

struct MailViewMutation
{
	MailViewMutation 		*Next;
	int						 AffectedRecord;
};

//==========================================================================================
// MailView
//
// A folder, in the GUI, represents a certain point-of-view towards the MailBox it is a part of. All
// messages that are received from the handler are children of the MailBox. The MailViews create
// MailViewNodes of any messages that are added to the MailBox if they are of relevance (which is,
// the filter rules match).
//==========================================================================================

class MailView
{
	public:
							 MailView (class MailBox *, char *);
							~MailView (void);
		
		MailView			*Next, *Prev;
		
		ASemaphore			*Lock;
		ASemaphore			*CacheLock;
		ASemaphore			*MutationLock;
		
		class MailBox		*Parent;
		char				*Name;
		
		int					 ProcessRules (class MailMessage *);
		void				 Associate (class MailMessage *);
		void				 Unassociate (class MailMessage *);

		int					 Count (void);
		MailMessage			*GetMessage (int);
		void				 DoneWithMessage (void);
		
		void				 ClearMutations (void);
		int				 	 GetNextMutation (void);
		
	private:
		int					_CachePos;
		MailViewNode		*Cache;
		
		MailViewNode		*First, *Last;
		int					_Count;
		MailViewMutation	*MFirst, *MLast;
};

//==========================================================================================
// ViewList/ViewListNode
//
// This is a child of the MailMessage. It is a pointer to a *subscriber* of the particular MailMessage.
// This class is the twin brother of the MailViewNode.
//==========================================================================================

class ViewListNode
{
	public:
						 ViewListNode (MailViewNode *);
						~ViewListNode (void);
		
		ViewListNode	*Next, *Prev;
		
		MailViewNode	*ViewNode;
		MailView		*View;
};

class ViewList
{
	public:
						 ViewList (MailMessage *);
						~ViewList (void);
		
		void			 Subscribe (MailViewNode *);
		void			 Unsubscribe (MailViewNode *);
		void			 UnTie (void);
		int				 Count (void);
		ViewListNode	*Get (int);
		
	private:
		MailMessage		*Parent;
		ViewListNode	*First, *Last;
		int				_Count;
};

//==========================================================================================
// MailMessage
//
// I'd dare say the most essential class of the bunch. This class represents a mailmessage as part of
// a mailbox. It has an envelope with one or more headers. It has one or more body-parts, or none in
// the case of a message that doesn't have a downloaded body yet.
//
// An MMFL_ORIGINAL flag keeps track of MailMessages that don't have an equivalent copy anymore on
// the server side. These should not get lost in garbage collection runs.
//
// The MMFL_HAVEBODY flag indicates whether the body of the message has been downloaded yet.
//
// MailParts may be flushed to disk at will (it is recommended to do this after a reasonable, but not
// too long, interval. 
//==========================================================================================

#define MMFL_ORIGINAL		 0x0001
#define MMFL_HAVEBODY		 0x0002
#define MMFL_INRAM			 0x0004
#define MMFL_ONDISK			 0x0008

class MailPart
{
	public:
						 MailPart (void);
						~MailPart (void);
		
		MailPart		*Next, *Prev;
		
		void			 Set (char *, char *, char *, unsigned long, char *);
		
		char			*ContentType;
		char			*FileName;
		char			*Description;
		unsigned long	 ContentSize;
		unsigned long	 pad1, pad2, pad3;
		char			*Contents;
};

class MailMessage
{
	public:
						 MailMessage (MailBox *, char *);
						~MailMessage (void);
		
		MailMessage		*Next, *Prev;
		
		ViewList		*Subscribers;
		WordList		*HeaderNames;
		WordList		*HeaderValues;
		WordList		*Envelope;
		
		void			 StoreFile (void);
		void			 Flush (void);
		void			 LoadBody (void);
		void			 SaveHeaders (void);
		void			 LoadHeaders (void);
		void			 RemoveFiles (void);
		
		FILE			*OpenFile (int, char *);
	
		char			*Header (char *);
			
		void			 ProcessMessage (char *);
		void			 ProcessHeaders (char *);
		void			 ProcessBody (char *);
		
		MailPart		*GetPart (int);
		void			 AddPart (char *, char *, char *, char *); // content-type, content-description, filename
		void			 AddPart (char *);
																											
		char			*MsgID;
		unsigned long	 cMsgID;
		
		MailBox			*Parent;
		
		int				 Count;
		unsigned int	 Flags;
		
	private:
		MailPart		*First, *Last;
		char			*MakeSafe (char *);
};

class MailHandler
{
	public:
		virtual					~MailHandler (void);
		virtual void			 OpenConnection (void);
		virtual void			 CloseConnection (void);
		virtual void			 GetBody (char *);
		virtual void			 CheckGroup (char *);
		virtual void			 DeleteMessage (char *);
		virtual void			 Expire (void);
		//virtual void			 MoveMessage (char *, char *, char *);

		char					 ConnectionOpen;
		Var						*Parameters;
		MailBox					*Parent;
};

class POP3Handler : public MailHandler
{
	public:
								 POP3Handler (MailBox *, Var *);
								~POP3Handler (void);

		void					 OpenConnection (void);
		void					 CloseConnection (void);
		void					 GetBody (char *);
		void					 CheckGroup (char *);
		void					 DeleteMessage (char *);
		void					 Expire (void);

		//void					 MoveMessage (char *, char *, char *);

	private:

		Var						*UIMsgID;
		Connection				*Con;
		int						 State;
		Var						*TransMsgID;
		int						 HighestHandled;

		int						 CheckReply (char *, int);
};

class MySQLHandler : public MailHandler
{
	public:
								 MySQLHandler (MailBox *, Var *);
								~MySQLHandler (void);

		void					 OpenConnection (void);
		void					 CloseConnection (void);
		void					 GetBody (char *);
		void					 CheckGroup (char *);
		void					 DeleteMessage (char *);
		void					 Expire (void);

	private:

		void					*SQL;
		int						 CheckError (void);
		int						 MsgToSQL (char *);
};

class NNTPHandler : public MailHandler
{
	public:
								 NNTPHandler (MailBox *, Var *);
								~NNTPHandler (void);

		void					 OpenConnection (void);
		void					 CloseConnection (void);
		void					 GetBody (char *);
		void					 CheckGroup (char *);
		void					 DeleteMessage (char *);
		void					 Expire (void);

	private:

		Connection				*Con;
		int						 CheckReply (int, char *, int);
		int						 CheckReply (int);
};

/*

class NNTPHandler : public MailHandler
{
	public:
								 NNTPHandler (MailBox *, Var *);
								~NNTPHandler (void);

		void					 OpenConnection (void);
		void					 CloseConnection (void);
		void					 GetBody (char *);
		void					 CheckGroup (char *);
		void					 DeleteMessage (char *);
		//void					 MoveMessage (char *, char *, char *);

	private:

		Connection				*Con;
		int						 State;
		Var						*TransMsgid;
		int						 HighestHandled;
};

*/

class MailBox
{
	public:
							 MailBox (ITCPipe *, char *);
							~MailBox (void);

		void				 Deliver (MailMessage *);
		void				 UnTie (MailMessage *);
		int					 Have (char *);

		void				 CreateMailView (char *);
		MailView			*GetView (char *);
		MailMessage			*Find (char *);
		void				 LoadMailBox();
		void				 SaveMailBox();
		void				 Quit();
		
		void				 RemoveMessage (char *);

		int				 	 Open;
		
		class MailHandler	*Handler;
		class MailerThread	*Collector;
		
		ITCPipe				*ToCollector;
		ITCPipe				*ToUI;
		Scheduler			*Schedule;

		unsigned long		 MemoryUsage;
		int					 Count;
		char				*Name;
		char				*DirectoryPrefix;
		int					 CacheToDisk;
				
		MailMessage			*First, *Last;
		MailView			*VFirst, *VLast;

		ASemaphore			*Lock;
		ASemaphore			*VLock;
};

#endif
