#include "mailbox.h"
#include "substr.h"
#include "eaddr.h"
#include "strdupp.h"
#include "chop.h"
#include "checksum.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

MailPart::MailPart (void)
{
	ContentType = NULL;
	FileName = NULL;
	Description = NULL;
	ContentSize = 0;
	Contents = NULL;
}

MailPart::~MailPart (void)
{
	if (ContentType ) delete[] (ContentType);
	if (FileName) delete[] (FileName);
	if (Description) delete[] (Description);
	if (Contents) delete[] (Contents);
}

void MailPart::Set (char *contenttype, char *filename, char *description, unsigned long contentsize, char *contents)
{
	ContentType = strdupp (contenttype);
	FileName = strdupp (filename);
	Description = strdupp (description);
	Contents = new char[contentsize+2];
	memcpy (Contents, contents, contentsize+1);
	//printf ("mailpart: set contents\n%s\n", Contents);
	ContentSize = contentsize;
}

MailMessage::MailMessage (MailBox *M, char *msgid)
{
	Subscribers = new ViewList (this);
	HeaderNames = new WordList;
	HeaderValues = new WordList;
	MsgID = strdupp (msgid);
	cMsgID = checksum (msgid);
	Flags = 0;
	Parent = M;
	Count = 0;
	Envelope = new WordList;
	First = Last = NULL;
	Next = Prev = NULL;
}

// Think straight for a bit here. Should the message bare responsibility for its contents or should this be handled by the
// mailbox it is a part of?
// The mailbox, hypothetically speaking, has a better way to know how to handle the in RAM / on disk cache, which means
// the only kind of flush we should do is make sure that message carrying MMFL_ORIGINAL do have a copy on disk before
// the message gets flushed.

MailMessage::~MailMessage (void)
{
	if ( (Flags & MMFL_ORIGINAL) && (Flags & MMFL_HAVEBODY) && (Flags & MMFL_INRAM) && (! Flags & MMFL_ONDISK) )
	{
		StoreFile();
	}
	delete (HeaderNames);
	delete (HeaderValues);
	delete (Subscribers);
	if (MsgID) delete (MsgID);
}


void MailMessage::RemoveFiles (void)
{
	int i;
	char *safe_msgid;
	char *fnam;
	safe_msgid = MakeSafe (MsgID);
	
	fnam = new char[ strlen (Parent->DirectoryPrefix) + 
					 strlen (safe_msgid) + 12 ];
	
	sprintf (fnam, "%s/%02x/%s.hd", Parent->DirectoryPrefix, cMsgID & 0xff, safe_msgid);
	unlink (fnam);
	
	if ((Flags & MMFL_ONDISK) && (Flags & MMFL_HAVEBODY))
	{
		for (i=0;i<Count;++i)
		{
			sprintf (fnam, "%s/%02x/%s.%02i", Parent->DirectoryPrefix, cMsgID & 0xff, safe_msgid, i);
			unlink (fnam);
		}
	}
	
	delete[] fnam;
	delete[] safe_msgid;
}

void MailMessage::StoreFile (void)
{
	FILE *X;
	int i,j;

	MailPart *C;
	
	if ((! (Flags & MMFL_ONDISK)) && (Flags & MMFL_HAVEBODY))
	{
		for (i=0;i<Count;++i)
		{
			C = GetPart (i);
			X = OpenFile (i, "w");
			if ((X)&&(C))
			{
				fprintf (X, "Content-Type: %s\n", C->ContentType);
				fprintf (X, "Content-Filename: %s\n", C->FileName);
				fprintf (X, "Content-Description: %s\n", C->Description);
				fprintf (X, "Content-Size: %d\n\n", C->ContentSize);
				fwrite (C->Contents, 1, C->ContentSize, X);
				fclose (X);
				Flags |= MMFL_ONDISK;
				Parent->SaveMailBox();
			}
			else
			{
				//if (!i) Flags &= (!MMFL_ONDISK);
				//j = i;
				i = Count;
				//Count = j;
			}
		}
	}
}

void MailMessage::SaveHeaders (void)
{
	FILE *X;
	char *fnam;
	int i;

	if (HeaderNames->Count)
	{
		X = OpenFile (-1, "w");
		if (X)
		{
			for (i=0;i<HeaderNames->Count;++i)
			{
				fprintf (X, "%s: %s\n", HeaderNames->Get(i), HeaderValues->Get(i));
			}
			fclose (X);
		}
	}
}

void MailMessage::LoadHeaders (void)
{
	FILE *X;
	char *fnam;
	char buf[512];
	char *headername;
	char *headervalue;
	char *boundary;

	X = OpenFile (-1, "r");
	if (X)
	{
		while (!feof (X))
		{
			buf[0] = 0;
			fgets (buf, 511, X);
			if ((buf[0]) && (!feof (X)))
			{
				choplf (buf);
				boundary = strchr (buf, ':');
				if (boundary)
				{
					headername = substr (buf, boundary);
					++boundary;
					while (*boundary == ' ') ++boundary;
					headervalue = boundary;
					HeaderNames->Add_ (headername);
					HeaderValues->Add (headervalue);
				}
			}
		}
		fclose (X);
	}
}

#define BADFILECHARS "!'`\"$&*?#%()[]\\\/\|{};:="

char *MailMessage::MakeSafe (char *_src)
{
	char *dst, *_dst;
	char *src = _src;
	
	_dst = new char[strlen (src)+2];
	dst = _dst;

	while (*src)
	{
		if (strchr (BADFILECHARS, *src)) *dst = '_';
		else if ((*src != '<') && (*src != '>'))
		{
			*dst = *src;
		}
		if ((*src == '<') || (*src == '>')) ++src;
		else
		{
			++src; ++dst;
		}
	}
	*dst = 0;
	return _dst;
}

FILE *MailMessage::OpenFile (int part, char *mode)
{
	FILE *result;

	char *fnam;
	char *safe_msgid;
	char *src, *dst;
	struct stat mystat;

	fnam = new char[strlen (Parent->DirectoryPrefix) + strlen (MsgID) + 32];

	sprintf (fnam, "%s/%02x", Parent->DirectoryPrefix, cMsgID & 0xff);
	if (stat (fnam, &mystat) != 0)
	{
		mkdir (fnam, 0700);
	}
	
	safe_msgid = MakeSafe (MsgID);
	if (part>=0) sprintf (fnam, "%s/%02x/%s.%02i", Parent->DirectoryPrefix, cMsgID & 0xff, safe_msgid, part);
	else sprintf (fnam, "%s/%02x/%s.hd", Parent->DirectoryPrefix, cMsgID & 0xff, safe_msgid);

	//printf ("fopen (\"%s\",\"%s\")\n", fnam, mode);

	result = fopen (fnam, mode);
	delete[] safe_msgid;
	delete[] fnam;
	return result;
}

// If there is body data in memory, save it to disk as needed and remove it from RAM

void MailMessage::Flush (void)
{
	int i;
	MailPart *P, *NP;
	
	if ( (Flags & MMFL_HAVEBODY) && (Flags & MMFL_INRAM) )
	{
		if (! (Flags & MMFL_ONDISK) )
		{
			StoreFile();
		}
		if (Flags & MMFL_ONDISK)
		{
			P = First;
			while (P)
			{
				NP = P->Next;
				Parent->MemoryUsage -= P->ContentSize;
				delete (P);
				P = NP;
			}
			First = Last = NULL;
			Flags &= (! MMFL_INRAM);
		}
	}
}

// If we have a body, but not in RAM, load the copy from disk

void MailMessage::LoadBody (void)
{
	//printf ("loadbody\n");
	if ( ! (Flags & MMFL_INRAM) )
	{
		//printf ("flags ok\n");
		int i;
		char buf[256];
		char contenttype[256];
		char description[256];
		char filename[256];
		unsigned long contentsize;
		char *v;
		int   done;
		
		int st;
		
		FILE *X;
		
		for (i=0;i<Count;++i)
		{
			//printf ("trying %i\n", i);
			X = OpenFile (i, "r");
			if (X)
			{
				st = 1;
				done = 0;
				while ((st)&&(!feof (X)))
				{ 
					buf[0] = 0;
					fgets (buf, 255, X);
					if (feof (X)) st = 0;
					else
					{
						choplf (buf);
						v = strchr (buf, ':');
						if (v)
						{
							while ( (*v==':') || (*v==' ') ) ++v;
							switch (st)
							{
								case 1:
									st = 2;
									strncpy (contenttype, v, 255); contenttype[255] = 0;
									break;
								case 2:
									st = 3;
									strncpy (filename, v, 255); filename[255] = 0;
									break;
								case 3:
									st = 4;
									strncpy (description, v, 255); description[255] = 0;
									break;
								case 4:
									st = 0;
									contentsize = atol (v);
									done = 1;
									break;
							}
						}
					}
				}
				if (done)
				{
					fgets (buf, 255, X);
					if (contentsize)
					{
						v = new char[contentsize+2];
						fread (v, 1, contentsize, X);
						v[contentsize] = 0;
						AddPart (contenttype, description, filename, v);// , contentsize);
						--Count;
					}
				}
			}
			if (X) fclose (X);
		}
		Flags |= MMFL_INRAM;
	}
}

void MailMessage::ProcessMessage (char *Message)
{
	char *headers;
	char *body;
	
	Count = 0;
	
	body = strstr (Message, "\n\n");
	if (!body)
	{
		ProcessHeaders (Message);
	}
	else
	{
		++body;
		if (body)
		{
			headers = substr (Message, body);
			++body;
			ProcessHeaders (headers);
			delete[] (headers);
			ProcessBody (body);
		}
	}
}

inline char *MailMessage::Header (char *name)
{
	int x = HeaderNames->Find (name);
	if (x>=0) return (HeaderValues->Get (x));
	
	return ((char *) "");
}

void MailMessage::ProcessHeaders (char *Hdr)
{
	if (!Hdr) return;
	int i;
	WordList X;
	char *lin;
	char *bnd;
	char *name;
	char *tmp;
	
	HeaderNames->Clear();
	HeaderValues->Clear();
	
	X.MakeGeneric (Hdr, '\n');
	for (i=0; i<X.Count; ++i)
	{
		lin = X.Get (i);
		if ((lin[0] == '\t') || (lin[0] == ' '))
		{
			while ((*lin == '\t')||(*lin == ' ')) ++lin;
			bnd = HeaderValues->Get (HeaderValues->Count -1);
			tmp = new char [strlen (lin) + strlen (bnd) + 3];
			strcpy (tmp, bnd);
			strcat (tmp, " ");
			strcat (tmp, lin);
			HeaderValues->Set (HeaderValues->Count -1, tmp);
			delete[] (tmp);
		}
		else if (bnd = strchr (lin, ':'))
		{
			name = substr (lin, bnd);
			while ( (*bnd == ':') || (*bnd == ' ') || (*bnd == '\t') ) ++bnd;
			HeaderNames->Add_ (name);
			HeaderValues->Add (bnd);
			//delete[] (name);
		}
	}
	//if (MsgID) delete (MsgID);
	
	//Envelope->Set (ENVELOPE_SUBJECT, Header ("Subject"));
	if (strlen (Header ("Newsgroups"))) Envelope->Set (ENVELOPE_NEWSGROUPS, Header ("Newsgroups"));
	Envelope->Set (ENVELOPE_DATE, Header ("Date"));
	tmp = eaddr2name (Header ("From"));
	Envelope->Set (ENVELOPE_NAMEFROM, tmp);
	delete[] (tmp);

	tmp = eaddr2addr (Header ("From"));
	if (tmp)
	{
		if (strlen (tmp) > 0) Envelope->Set (ENVELOPE_MAILFROM, tmp);
		delete[] (tmp);
	}
	Envelope->Set (ENVELOPE_SUBJECT, Header ("Subject"));
	SaveHeaders();
}

void MailMessage::ProcessBody (char *body)
{
	int	  didmultipart = 0;
	char *contenttype;
	char *boundary;
	char *mimetype;
	
	Count = 0;
	
	contenttype = Header ("Content-Type"); // Get the Content-Type header
	if (*contenttype)
	{
		printf ("<--- Content-Type: %s --->\n", contenttype);
		WordList CT;
		CT.MakeGeneric (contenttype, ';'); // Seperate arguments
		mimetype = strdupp (CT.Get(0));
		chopwhitespace (mimetype);
		if ( (strcasecmp (mimetype, "multipart/alternative") == 0) || // a.k.a. OutLook HTML Cannon
			 (strcasecmp (mimetype, "multipart/mixed") == 0) )
		{
			printf ("Korben Dallas MULTIPART\n");
			for (int i = 1; i<CT.Count;++i)
			{
				WordList Sub;
				char *subname;
				char *subval;
				char *tmp;
				
				Sub.MakeGeneric (CT.Get(i), '='); // Split parameters
				subname = Sub.Get (0);
				chopwhitespace (subname);
				printf ("checking subname: '%s'\n", subname);
				if (strcasecmp (subname, "boundary") == 0) // boundary parameter?
				{
					subval = Sub.Get (1);
					while ( (*subval == ' ') || (*subval == '\t') ) ++subval; // find value
					chopwhitespace (subval);
					if (subval[0] == '\"') // value between double quotes
					{
						++subval;
						tmp = strchr (subval, '\"'); // find second quote
						if (tmp) *tmp = 0;
					}
					else if (subval[0] == '\'') // value between single quotes
					{
						++subval;
						tmp = strchr (subval, '\''); // find second quote
						if (tmp) *tmp = 0;
					}
					boundary = strdupp (subval); // value unquoted
					printf ("boundary=%s\n", boundary);
					if (strlen (boundary) > 3)
					{
						char *partstart;
						char *partend;
						char *part;
						
						didmultipart = 1;
						partstart = strstr (body, boundary); // find first part
						while (partstart)
						{
							partstart += strlen (boundary);
							while (*partstart < 32) ++partstart;
							partend = strstr (partstart, boundary); // find end of part
							if (partend)
							{
								part = substr (partstart, partend);
								AddPart (part);
								delete[] (part);
							}
							partstart = partend;
						}
					}
					delete[] (boundary);
				}
			}
		}
		delete[] (mimetype);
	}
	if (!didmultipart) // This wasn't a multipart message I guess, well, add it all as text/plain then
	{
		//printf ("plaintext single-part message:\n%s\n", body);
		AddPart ("text/plain","","", body);
	}

	Flags |= MMFL_HAVEBODY;
	
	StoreFile();
}

void MailMessage::AddPart (char *contenttype, char *description, char *filename, char *body)
{
	MailPart *M = new MailPart;
	M->Set (contenttype, filename, description, strlen(body), body);
	if (First)
	{
		M->Next = NULL;
		M->Prev = Last;
		Last->Next = M;
		Last = M;
	}
	else
	{
		M->Prev = M->Next = NULL;
		First = Last = M;
	}
	++Count;
}

// Extract headers form a multipart element and add it as a MailPart.

MailPart *MailMessage::GetPart (int cnt)
{
	MailPart *P = First;
	int i = 0;
	while ((P) && (i<cnt))
	{
		P = P->Next;
		++i;
	}
	return P;
}

void MailMessage::AddPart (char *part)
{
	WordList W;
	char *ln;
	char *oln;
	char *end;
	char *val;
	char *body;
	int     i;
	char contenttype[256];
	char description[256];
	char filename[256];
	WordList HNames, HValues;
	
	contenttype[0] = 0;
	description[0] = 0;
	filename[0] = 0;
	
	end = strchr (part, '\n');
	if (end) ln = substr (part, end);
	while (ln && (strlen (ln) > 1))
	{
		char *colonpos;
		char *hdrname;
		char *hdrvalue;
		char *crsr = ln;

		if ( ((ln[0] == ' ') || (ln[0] == '\t')) && HValues.Count)
		{
			while ( (*crsr==' ') || (*crsr=='\t') ) ++crsr;
			hdrvalue = new char [strlen (HValues.Get(HValues.Count -1)) + strlen (crsr) + 1];
			strcpy (hdrvalue, HValues.Get (HValues.Count -1));
			strcat (hdrvalue, crsr);
			HValues.Set (HValues.Count-1, hdrvalue);
			delete (hdrvalue);
		}
		else
		{
			colonpos = strchr (ln, ':');
			if (colonpos)
			{
				hdrname = substr (ln, colonpos);
				hdrvalue = colonpos+1;
				while ( (*hdrvalue==' ') || (*hdrvalue=='\t') ) ++hdrvalue;
				HNames.Add_ (hdrname);
				HValues.Add (hdrvalue);
			}
		}
		
		crsr = end+1;
		delete[] (ln);
		end = strchr (crsr, '\n');
		if (end)
		{
			ln = substr (crsr,end);
		}
		else ln = NULL;
	}
	if (ln) delete[] (ln);
	body = end+1;
	
	i = HNames.Find ("Content-Type");
	if (i>=0)
	{
		strncpy (contenttype, HValues.Get(i), 255);
		contenttype[255] = 0;
	}
	i = HNames.Find ("Content-Description");
	if (i>=0)
	{
		strncpy (description, HValues.Get(i), 255);
		description[255] = 0;
	}
	AddPart (contenttype, description, filename, body);
	HNames.Clear();
	HValues.Clear();
}
