//===============================================================================
// POSTOFFICE - X11 Office Communication Handler
// ---------------------------------------------
// All sourcecode, unless differently annotated, written by Pim van Riezen
// Sourcecode and resulting binaries are covered by the GNU Lesser General Public
// License (GNU LGPL). For more information and the literal text of the license
// agreement, check the Free Software Foundation's website on http://www.fsf.org/
// 
// Copyright (C) 1999 Mad Sciene Labs - Rotterdam, The Netherlands
//===============================================================================

#include "mailstorage.h"
#include "mailbox.h"
#include "strdupp.h"
#include "debug.h"
#include "gdbmtools.h"

#include <gdbm.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>

GDBMStorage::GDBMStorage (const char *filename)// : MailStorage()
{
	Filename = Name = strdupp (filename);
	DB = gdbm_open (Filename, 0, GDBM_WRCREAT | GDBM_FAST, 0600, NULL);
	if (!DB)
	{
		eprintf ("gdbmstorage: could not open: %s\n", gdbm_strerror (gdbm_errno));
	}
	lastsync = time (NULL);
	BF = new AsyncBuffer (4096);
    //Root = NULL;
	Root = new StorageNode (this, NULL, 0, 0);
	SFirst = SLast = NULL;
	Size_ = 0;
	Count_ = 0;
	Limit_ = 0;
	SCount_ = 0;
	Lock = new ASemaphore ("mailstorage");
}

GDBMStorage::~GDBMStorage (void)
{
	StoreIndex(true);
	if (DB) gdbm_close(DB);
	dprintf ("gdbmstorage [%s] deleted\n", Filename);
	delete[] Filename;
	delete (BF);
}

void GDBMStorage::Store (MailMessage *M)
{
	if (DB)
	{
		Lock->LockOnce();
		StorageNode *N = Find_ (M->MsgID);
		
		char *key = new char[strlen (M->MsgID) + 16];
		datum mydat, keydat;
		
		bool stored_data = false;
		int total_size = 0;
		int partcount = 0;
		
		// If there are any real headers, store them
		
		if (!N && (M->HeaderNames->Count))
		{
			BF->Pos = 0;
			
			dprintf ("gdbmstorage: storehdr %s\n", M->MsgID);
			
			sprintf (key, "%s/hdr", M->MsgID);
			keydat.dptr = key;
			keydat.dsize = strlen (key);
			
			for (int i=0; i<M->HeaderNames->Count;++i)
			{
				BF->Add (M->HeaderNames->Get(i));
				BF->Add ((unsigned char) '\n');
				BF->Add (M->HeaderValues->Get(i));
				BF->Add ((unsigned char) '\n');
			}
			BF->Add ((unsigned char) 0);
			
			mydat.dptr = (char *) BF->Data;
			mydat.dsize = BF->Pos;
			
			total_size += BF->Pos;
			
			gdbm_store (DB, keydat, mydat, GDBM_INSERT);
			stored_data = true;
		}
		
		// If the message contents are in RAM, store them
		
		if ((N && (N->Count < M->Count)) && ((M->Flags & (MMFL_HAVEBODY | MMFL_INRAM))))
		{
			dprintf ("gdbmstorage: store %s\n", M->MsgID);
			MailPart *C;
			
			// Iterate all parts
			
			for (int i=0; i<M->Count;++i)
			{
				keydat.dptr = key;
				
				if ((C = M->GetPart (i))) // Part exists?
				{
					++partcount;
					
					sprintf (key, "%s.%02i/dat", M->MsgID, i);
					keydat.dsize = strlen (key);
					
					int sz = C->ContentSize +1;
					
					char *dat = gdbm_createnode (C->ContentType, C->FileName, C->Description, C->Contents, &sz);
					
					mydat.dptr = dat;
					mydat.dsize = sz;
					
					total_size += sz;
					
					gdbm_store (DB, keydat, mydat, GDBM_INSERT);
					stored_data = true;
					delete[] dat;
					dprintf ("gdbmstorage: store (%s)\n", keydat.dptr);
				}
			}
		}
		else
		{
			dprintf ("gdbmstorage: cannot save empty message %s\n", M->MsgID);
		}
		
		if (stored_data)
		{
			//gdbm_sync (DB);
			Add (M->MsgID, (size_t) total_size, partcount);
		}
		
		Lock->UnlockOnce();
		
		delete[] key;
	}
	else eprintf ("gdbmstorage: GDBM not open\n");
}

bool GDBMStorage::Retrieve (MailMessage *M)
{
	if (!DB)
	{
		eprintf ("gdbmstorage: GDBM db not open\n");
		return false;
	}
	
	bool result = true;
	
	Lock->LockOnce();
	StorageNode *Nod = Find_ (M->MsgID);
	
	if (Nod)
	{
		char *key = new char[strlen (M->MsgID) + 16];
		datum mydat, keydat;
		
		if (!M->HeaderNames->Count)
		{
			WordList WL;
			
			sprintf (key, "%s/hdr", M->MsgID);
			keydat.dptr = key;
			keydat.dsize = strlen (key);
			
			mydat = gdbm_fetch (DB, keydat);
			if (mydat.dptr)
			{
				WL.MakeGeneric (mydat.dptr, '\n', 1);
				for (int i=0; (i+1)<WL.Count; i+=2)
				{
					M->HeaderNames->Add (WL.Get(i));
					M->HeaderValues->Add (WL.Get(i+1));
				}
				free (mydat.dptr);
				result = true;
			}
			else
			{
				eprintf ("[gdbmstorage] (%s) gdbm error: %s\n", key, gdbm_strerror (gdbm_errno));
				result = false;
			}
		}
		if ( (!(M->Flags & MMFL_INRAM)) && (M->Flags & MMFL_HAVEBODY) )
		{
			char *filename;
			char *contenttype;
			char *description;
			char *nc;
			
			iprintf ("[gdbmstorage] downloading %i parts\n", M->Count);
			
			for (int i=0; i<M->Count; ++i)
			{
				sprintf (key, "%s.%02i/dat", M->MsgID, i);
				keydat.dptr = key;
				keydat.dsize = strlen (key);
				
				mydat = gdbm_fetch (DB, keydat);
				if (mydat.dptr)
				{
					contenttype = gdbm_node2contenttype (mydat.dptr, mydat.dsize);
					filename = gdbm_node2filename (mydat.dptr, mydat.dsize);
					description = gdbm_node2description (mydat.dptr, mydat.dsize);
					nc = gdbm_nodecontents (mydat.dptr, mydat.dsize);
					
					MailPart *P = M->GetPart (i);
					if (P)
					{
						P->Set (contenttype, filename, description, (unsigned long) strlen (nc), nc);
					}
					else
					{
						M->AddPart (contenttype, description, filename, nc);
						M->Count--;
					}
					
					free (mydat.dptr);
					//delete[] nc;
					delete[] description;
					delete[] contenttype;
					delete[] filename;
					result = true;
				}
				else
				{
					eprintf ("[gdbmstorage] (%s) gdbm error: %s\n", key, gdbm_strerror (gdbm_errno));
					result = false;
				}
			}
		}
		delete[] key;
	}
	Lock->UnlockOnce();
	return result;
}

void GDBMStorage::Delete (const char *key)
{
	if (!DB)
	{
		eprintf ("gdbmstorage: GDBM db not open\n");
		return;
	}
	
	Lock->LockOnce();
	
	StorageNode *N = Find (key);
	Delete (N);
	Lock->UnlockOnce();
}

void GDBMStorage::Delete (StorageNode *N)
{
	if (!DB)
	{
		eprintf ("gdbmstorage: GDBM db not open\n");
		return;
	}
	
	if (N)
	{
		char *key = new char[strlen (N->BKey) + 16];
		datum keydat;
		
		sprintf (key, "%s/hdr", N->BKey);
		keydat.dptr = key;
		keydat.dsize = strlen(key);
		gdbm_delete (DB, keydat);
		
		dprintf ("gdbmstorage: deleted (%s)\n", key);
		
		for (int i=0; i<N->Count; ++i)
		{
			sprintf (key, "%s.%02i/dat", N->BKey, i);
			keydat.dsize = strlen(key);
			gdbm_delete (DB, keydat);
			dprintf ("gdbmstorage: deleted (%s)\n", key);
		}
		
		gdbm_sync (DB);
		
		delete[] key;
		Remove (N);
	}
	else
	{
		eprintf ("gdbmstorage: delete: does not exist\n");
	}
}

void GDBMStorage::StoreIndex (bool force)
{
	if (!DB)
	{
		eprintf ("gdbmstorage: GDBM not open\n");
		return;
	}
	
	if ((!force) && ((time (NULL) - lastsync)<15))
	{
		return;
	}
	
	lastsync = time (NULL);
	
	BF->Pos = 0;
//	char buf[32];
	datum keydat, mydat;
	
	if (Root->BSame) StoreNode ((StorageNode *)Root->BSame);
	if (Root->BHigher) StoreNode ((StorageNode *)Root->BHigher);
	if (Root->BLower) StoreNode ((StorageNode *)Root->BLower);
	
	BF->Add ((unsigned char) 0);
	
	int padding = 4096 - (BF->Pos & 4095);
	if (padding>0) BF->Grow (padding);
	
	keydat.dptr = "index";
	keydat.dsize = 5;
	mydat.dptr = (char *) BF->Data;
	mydat.dsize = BF->Pos;
	
	gdbm_store (DB, keydat, mydat, GDBM_REPLACE);
	gdbm_sync (DB);
}

void GDBMStorage::LoadIndex (void)
{
	if (!DB)
	{
		eprintf ("gdbmstorage: GDBM not open\n");
		return;
	}
	
	WordList WL;
	WordList WWL;
	datum keydat, mydat;
	
	keydat.dptr = "index";
	keydat.dsize = 5;
	
	Lock->LockOnce();
	
	mydat = gdbm_fetch (DB, keydat);
	if (mydat.dptr)
	{
		WL.MakeGeneric (mydat.dptr, '\n');
		for (int i=0; i<WL.Count; ++i)
		{
			WWL.MakeGeneric (WL.Get(i), '\t');
			//dprintf ("gdbmstorage: inserting key: %s\n", WWL.Get(0));
			Add (WWL.Get(0), (size_t) atoi (WWL.Get(1)), atoi (WWL.Get(2)));
		}
		iprintf ("\n");
	}
	else
	{
		iprintf ("gdbmstorage: no index segment found\n");
	}
	Lock->UnlockOnce();
}

void GDBMStorage::StoreNode (StorageNode *N)
{
	char buf[32];
	
	
	
	if (N)
	{
		BF->Add (N->BKey);
		BF->Add ((unsigned char) '\t');
		
		sprintf (buf, "%i", N->Size);
		BF->Add (buf);
		BF->Add ((unsigned char) '\t');
		
		sprintf (buf, "%i", N->Count);
		BF->Add (buf);
		BF->Add ((unsigned char) '\n');
		
		if (N->BSame) StoreNode ((StorageNode *)N->BSame);
		if (N->BHigher) StoreNode ((StorageNode *)N->BHigher);
		if (N->BLower) StoreNode ((StorageNode *)N->BLower);
	}
}
