//===============================================================================
// POSTOFFICE - X11 Office Communication Handler
// ---------------------------------------------
// All sourcecode, unless differently annotated, written by Pim van Riezen
// Sourcecode and resulting binaries are covered by the GNU Lesser General Public
// License (GNU LGPL). For more information and the literal text of the license
// agreement, check the Free Software Foundation's website on http://www.fsf.org/
// 
// Copyright (C) 1999 Mad Sciene Labs - Rotterdam, The Netherlands
//===============================================================================

#include "mailstorage.h"
#include "mailbox.h"
#include "debug.h"
#include "checksum.h"

StorageList::StorageList (void)
{
	First = Last = NULL;
	Lock = new ASemaphore ("storagelist");
}

StorageList::~StorageList (void)
{
	if (First) eprintf ("storagelist still had subscribers on termination!\n");
	delete (Lock);
}

MailStorage *StorageList::Open (const char *name, class MailBox *subs, short Type)
{
	Lock->LockOnce();
	
	MailStorage *N = Find (name);
	if (!N)
	{
		switch (Type)
		{
			case STORAGE_GDBM:
				N = new GDBMStorage (name);
				N->LoadIndex ();
				N->Subscribe (subs);
				break;

			case STORAGE_MBOX:
				N = new MBoxStorage (name);
				N->LoadIndex ();
				N->Subscribe (subs);
				break;

			default:
				eprintf ("unsupported storage class: %02x\n", Type);
				exit (1);
				break;
		}
		if (N)
		{
			if (Last)
			{
				N->Prev = Last;
				N->Next = NULL;
				Last->Next = N;
				Last = N;
			}
			else
			{
				N->Prev = N->Next = NULL;
				First = Last = N;
			}
		}
	}
	else
	{
		N->Subscribe (subs);
	}

	Lock->UnlockOnce();
	return N;
}

void StorageList::Close (MailBox *M)
{
	Lock->LockOnce();
		MailStorage *N = First;
		MailStorage *NN;
		while (N)
		{
			NN = N->Next;
			N->Unsubscribe (M);
			
			if (!N->SCount()) // No subscribers left
			{
				if (N->Prev)
				{
					if (N->Next)
					{
						N->Prev->Next = N->Next;
						N->Next->Prev = N->Prev;
					}
					else
					{
						N->Prev->Next = NULL;
						Last = N->Prev;
					}
				}
				else
				{
					if (N->Next)
					{
						N->Next->Prev = NULL;
						First = N->Next;
					}
					else
					{
						First = Last = NULL;
					}
				}
				delete (N);
			}
			N = NN;
		}
	Lock->UnlockOnce();
}

MailStorage *StorageList::Find (const char *nm)
{
	MailStorage		*result = NULL;
	MailStorage		*N;

		N = First;
		while (N)
		{
			if ( strcmp (N->Name, nm) == 0)
			{
				result = N;
				N = NULL;
			}
			else N=N->Next;
		}

	return result;
}
